--[[ Copyright 2008 Erich Blume
    This file is part of Socialist.

    Socialist is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    Socialist is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with Socialist.  If not, see <http://www.gnu.org/licenses/>.
]]--


assert(Socialist,"Could not find Socialist.")
---------------------------------------
-- Variable Definitions              --
---------------------------------------	
SocialistUI = Socialist:NewModule("SocialistUI",
	"AceConsole-3.0"
	)
	
local wholib = LibStub:GetLibrary('LibWho-2.0'):Library()


local options = {
	name = "SocialistUI",
	handler = SocialistUI,
	type = "execute",
	desc = "Toggle the default Socialust user interface.",
	func = "ToggleMainForm",
}

local _G = getfenv()
local sf = CreateFrame("Frame",nil,UIParent)
SocialistUI.socialistForm = sf

---------------------------------
-- EVENTS                      --
---------------------------------

function SocialistUI:OnInitialize()
	LibStub("AceConfig-3.0"):RegisterOptionsTable("SocialistUI",options,{"socialistui"})
end


---------------------------------
-- CONSTANTS                   --
---------------------------------

local TITLE = "Socialist v0.1 BETA"

-- Number of entries to show at once. Changing this will break the layout.
-- TODO: Replace this with resizing code.
local MAXENTRIES = 17

-- Seconds between processing the input text from the form.
-- TODO: Test this in real conditions. Initial tests show even setting this to 0
--		 isn't enough to cause problems.
local MIN_TEXT_UPDATE_INTERVAL = 0.1

-- Level informaton. Unknown Level is a special value, -1
local UNKNOWN_LEVEL = -1
local MAX_LEVEL = 255 -- GMs can be level 255

---------------------------------------
-- Local Hooks                       --
---------------------------------------
-- This somewhat senseless section is for optimization. Ignore it.
local tremove = _G["tremove"]
local tinsert = _G["tinsert"]

---------------------------------
-- VIRTUALS                    --
---------------------------------
local newResultEntry = function(parent)
	local result = CreateFrame("Button",nil,parent)
	result:SetHeight(17)
	result:SetHighlightTexture("Interface\\QuestFrame\\UI-QuestTitleHighlight","ADD")
	result:RegisterForClicks("AnyUp")
	result:SetScript("OnClick",function(self, button, down) SocialistUI:ResultClick(self) end)

	local name = CreateFrame("Frame",nil,result)
	name:SetPoint("TOPLEFT"); name:SetPoint("BOTTOMLEFT")
	name:SetWidth(120)

	local nt = name:CreateFontString()
	nt:SetFontObject(GameFontNormal)
	nt:SetText("")
	nt:SetAllPoints(name)
	nt:SetJustifyH("LEFT")
	nt:SetTextColor(1,0.8196079,0)
	name.text = nt

	local guild = CreateFrame("Frame",nil,result)
	guild:SetPoint("TOPLEFT",name,"TOPRIGHT")
	guild:SetPoint("BOTTOMLEFT",name,"BOTTOMRIGHT")
	guild:SetWidth(120)

	local gt = guild:CreateFontString()
	gt:SetFontObject(GameFontNormal)
	gt:SetText("")
	gt:SetAllPoints(guild)
	gt:SetJustifyH("LEFT")
	gt:SetTextColor(1,0.8196079,0)
	guild.text = gt

	local level = CreateFrame("Frame",nil,result)
	level:SetPoint("TOPLEFT",guild,"TOPRIGHT")
	level:SetPoint("BOTTOMLEFT",guild,"BOTTOMRIGHT")
	level:SetWidth(38)

	local lt = level:CreateFontString()
	lt:SetFontObject(GameFontNormal)
	lt:SetText("")
	lt:SetAllPoints(level)
	lt:SetJustifyH("LEFT")
	lt:SetTextColor(1,0.8196079,0)
	level.text = lt

	local class = CreateFrame("Frame",nil,result)
	class:SetPoint("TOPLEFT",level,"TOPRIGHT")
	class:SetPoint("BOTTOMLEFT",level,"BOTTOMRIGHT")
	class:SetWidth(75)

	local ct = class:CreateFontString()
	ct:SetFontObject(GameFontNormal)
	ct:SetText("")
	ct:SetAllPoints(class)
	ct:SetJustifyH("LEFT")
	ct:SetTextColor(1,0.8196079,0)
	class.text = ct

	result.name = name
	result.guild = guild
	result.level = level
	result.class = class
	return result
end

---------------------------------
-- SocialistForm              --
---------------------------------
sf:EnableMouse(true)
sf:SetMovable(true)
sf:Hide()
sf:SetWidth(400)
sf:SetHeight(500)
sf:SetPoint("CENTER")
sf:SetBackdrop({
	bgFile = "Interface\\DialogFrame\\UI-DialogBox-Background",
	edgeFile="Interface\\DialogFrame\\UI-DialogBox-Border",
	tile= true,
	insets = {left=11, right=12, top=12, bottom=11},
	tileSize = 32,
	edgeSize = 32,
	})

sf:SetScript("OnMouseUp", function(frame)
	if ( frame.isMoving ) then
		frame:StopMovingOrSizing()
		frame.isMoving = false
	end
end)
sf:SetScript("OnMouseDown", function(frame,arg1)
	if (((not frame.isLocked)or(frame.isLocked == 0))and(arg1 == "LeftButton")) then
		frame:StartMoving()
		frame.isMoving = true
	end
end)

---- TITLE
-- TODO: find a good FontString to inherit from.
do
	local title = sf:CreateFontString()
	title:SetFontObject(GameFontNormal)
	title:SetText(TITLE)
	title:SetPoint("TOP",sf,"TOP",0,-10)
	title:SetTextColor(1,0.8196079,0)
	-- title:SetWidth(140); title:SetHeight(0); -- Removed because I don't see why it's needed.
	-- Also removed some shadow stuff... I suspect it unnecessary as well.
	sf.titleFrame = title
end

---- CLOSE
do
	local button = CreateFrame("Button",nil,sf,"UIPanelCloseButton")
	button:SetPoint("TOPRIGHT",sf,"TOPRIGHT",-5,-5)
	sf.closeButton = button
end


---------------------------------
-- FilterBox                   --
---------------------------------
do
	local fb = CreateFrame("Frame",nil,sf)
	fb:SetPoint("TOPLEFT",sf,"TOPLEFT",10,-30)
	fb:SetPoint("TOPRIGHT",sf,"TOPRIGHT",-10,-30)
	fb:SetHeight(150)
	fb:SetBackdrop({
		bgFile="Interface\\Tooltips\\UI-Tooltip-Background",
                edgeFile="Interface\\Tooltips\\UI-Tooltip-Border",
                tile = true,
                tileSize = 16,
                edgeSize = 16,
                insets = {left=4, right=4, top=4, bottom=4},
		})
	do	-- Name Search Box
		local topoffset = -1
		local eb = CreateFrame("EditBox",nil,fb)
		eb:SetFontObject("ChatFontNormal")
		eb:EnableMouse(true)
		eb:SetWidth(120); eb:SetHeight(30);
		eb:SetPoint("TOPLEFT",fb,"TOPLEFT",75,topoffset)
		eb:SetScript("OnTextChanged",function() SocialistUI:TextChanged() end)
		eb:SetScript("OnEnterPressed",function() SocialistUI:TextChanged() end)
		local title = eb:CreateFontString()
		title:SetFontObject(GameFontNormal)
		title:SetText("Name")
		title:SetJustifyH("RIGHT")
		title:SetJustifyV("BOTTOM")
		title:SetTextColor(1,0.8196079,0)
		title:SetPoint("TOPLEFT",fb,"TOPLEFT",6,topoffset-10)
		local tex = eb:CreateTexture()
		tex:SetPoint("TOPLEFT",eb,"TOPLEFT",-2,-5)
		tex:SetPoint("TOPRIGHT",eb,"TOPRIGHT",0,-5)
		tex:SetPoint("BOTTOMLEFT",eb,"BOTTOMLEFT",-2,-5)
		tex:SetPoint("BOTTOMRIGHT",eb,"BOTTOMRIGHT",0,-5)
		tex:SetTexture("Interface\\Common\\Common-Input-Border")
		eb.texture = tex
		fb.nameBox = eb
	end
	
	do -- Guild Search Box
		local topoffset = -22
		local eb = CreateFrame("EditBox",nil,fb)
		eb:SetFontObject("ChatFontNormal")
		eb:EnableMouse(true)
		eb:SetWidth(120); eb:SetHeight(30);
		eb:SetPoint("TOPLEFT",fb,"TOPLEFT",75,topoffset)
		eb:SetScript("OnTextChanged",function() SocialistUI:TextChanged() end)
		eb:SetScript("OnEnterPressed",function() SocialistUI:TextChanged() end)
		local title = eb:CreateFontString()
		title:SetFontObject(GameFontNormal)
		title:SetText("Guild")
		title:SetJustifyH("RIGHT")
		title:SetJustifyV("BOTTOM")
		title:SetTextColor(1,0.8196079,0)
		title:SetPoint("TOPLEFT",fb,"TOPLEFT",6,topoffset-10)
		local tex = eb:CreateTexture()
		tex:SetPoint("TOPLEFT",eb,"TOPLEFT",-2,-5)
		tex:SetPoint("TOPRIGHT",eb,"TOPRIGHT",0,-5)
		tex:SetPoint("BOTTOMLEFT",eb,"BOTTOMLEFT",-2,-5)
		tex:SetPoint("BOTTOMRIGHT",eb,"BOTTOMRIGHT",0,-5)
		tex:SetTexture("Interface\\Common\\Common-Input-Border")
		eb.texture = tex
		fb.guildBox = eb
	end
	
	do -- Level search Box
		local topoffset = -43
		local eba = CreateFrame("EditBox",nil,fb)
		eba:SetFontObject("ChatFontNormal")
		eba:EnableMouse(true)
		eba:SetWidth(30); eba:SetHeight(30);
		eba:SetPoint("TOPLEFT",fb,"TOPLEFT",75,topoffset)
		eba:SetScript("OnTextChanged",function() SocialistUI:TextChanged() end)
		eba:SetScript("OnEnterPressed",function() SocialistUI:TextChanged() end)
		local titlea = eba:CreateFontString()
		titlea:SetFontObject(GameFontNormal)
		titlea:SetText("Level")
		titlea:SetJustifyH("RIGHT")
		titlea:SetJustifyV("BOTTOM")
		titlea:SetTextColor(1,0.8196079,0)
		titlea:SetPoint("TOPLEFT",fb,"TOPLEFT",6,topoffset-10)
		local texa = eba:CreateTexture()
		texa:SetPoint("TOPLEFT",eba,"TOPLEFT",-2,-5)
		texa:SetPoint("TOPRIGHT",eba,"TOPRIGHT",0,-5)
		texa:SetPoint("BOTTOMLEFT",eba,"BOTTOMLEFT",-2,-5)
		texa:SetPoint("BOTTOMRIGHT",eba,"BOTTOMRIGHT",0,-5)
		texa:SetTexture("Interface\\Common\\Common-Input-Border")
		eba.texture = texa
		fb.levelMinBox = eba
		
		local ebb = CreateFrame("EditBox",nil,fb)
		ebb:SetFontObject("ChatFontNormal")
		ebb:EnableMouse(true)
		ebb:SetWidth(30); ebb:SetHeight(30);
		local titleb = ebb:CreateFontString()
		titleb:SetFontObject(GameFontNormal)
		titleb:SetText("to")
		titleb:SetJustifyH("CENTER")
		titleb:SetJustifyV("BOTTOM")
		titleb:SetTextColor(1,0.8196079,0)
		titleb:SetPoint("LEFT",eba,"RIGHT",6,0)
		ebb:SetPoint("LEFT",titleb,"RIGHT",6,0)
		ebb:SetScript("OnTextChanged",function() SocialistUI:TextChanged() end)
		ebb:SetScript("OnEnterPressed",function() SocialistUI:TextChanged() end)
		local texb = ebb:CreateTexture()
		texb:SetPoint("TOPLEFT",ebb,"TOPLEFT",-2,-5)
		texb:SetPoint("TOPRIGHT",ebb,"TOPRIGHT",0,-5)
		texb:SetPoint("BOTTOMLEFT",ebb,"BOTTOMLEFT",-2,-5)
		texb:SetPoint("BOTTOMRIGHT",ebb,"BOTTOMRIGHT",0,-5)
		texb:SetTexture("Interface\\Common\\Common-Input-Border")
		ebb.texture = texb
		fb.levelMaxBox = ebb
	end
	
	do -- Update Button
		local ub = CreateFrame("Button",nil,fb,"OptionsButtonTemplate")
		ub:SetPoint("BOTTOMLEFT",fb,"BOTTOMLEFT",4,6);
		ub:SetWidth(120)
		ub:SetText("Update List")
		ub:RegisterForClicks("LeftButtonUp")
		ub:SetScript("OnClick",function(self,button,down) SocialistUI:UpdateList() end)

		fb.updateListButton = ub
	end
	
	do -- Query Button
		local qb = CreateFrame("Button",nil,fb,"OptionsButtonTemplate")
		qb:SetPoint("BOTTOMRIGHT",fb,"BOTTOMRIGHT",-4,6);
		qb:SetWidth(120)
		qb:SetText("Query Server")
		qb:RegisterForClicks("LeftButtonUp")
		qb:SetScript("OnClick",function(self,button,down) SocialistUI:QueryServer() end)

		fb.queryButton = qb
	end
	
	
	sf.filterBox = fb
end



---------------------------------
-- ResultBox                   --
---------------------------------
do
	local rb = CreateFrame("Frame",nil,sf)
	rb:SetPoint("BOTTOMLEFT",sf,"BOTTOMLEFT",10,10)
	rb:SetPoint("BOTTOMRIGHT",sf,"BOTTOMRIGHT",-10,10)
	rb:SetHeight(312)
	rb:SetBackdrop({
		bgFile="Interface\\Tooltips\\ChatBubble-Background",
                edgeFile="Interface\\Tooltips\\UI-Tooltip-Border",
                tile = true,
                tileSize = 16,
                edgeSize = 16,
                insets = {left=4, right=4, top=4, bottom=4},
		})
	rb:EnableMouseWheel(1)
	-- OMW funtion will call ScrollBar, ok because this won't be called till later
	rb:SetScript("OnMouseWheel",function(frame,delta)
		local sb = sf.resultBox.scrollBar
		sb:SetValue(sb:GetValue() - delta * sb:GetValueStep())
		end)
		
	-- Scrollbar
	local sb = CreateFrame("Slider",nil,rb)
	sb:SetPoint("TOPRIGHT",rb,"TOPRIGHT",-6,-16)
	sb:SetPoint("BOTTOMRIGHT",rb,"BOTTOMRIGHT",-6,13)
	sb:SetOrientation("VERTICAL")
	sb:SetWidth(16)
	sb:SetMinMaxValues(1,14) -- defaults
	sb:SetValue(1) -- defaults
	sb:SetValueStep(1)
	sb:SetScript("OnValueChanged",function(frame,value) 
		SocialistUI.topResultIndex = value
		SocialistUI:RedrawList() 
		end)
	sb:SetThumbTexture("Interface\\Buttons\\UI-ScrollBar-Knob")
	
	---- up arrow
	local up = CreateFrame("Button",nil,sb,"UIPanelScrollUpButtonTemplate")
	up:SetPoint("BOTTOM",sb,"TOP",0,-7)
	up:SetScript("OnClick",function(self,button,down)
		PlaySound("UChatScrollButton")
		sb:SetValue(sb:GetValue() - sb:GetValueStep())
	end)
	
	---- down arrow
	local down = CreateFrame("Button",nil,sb,"UIPanelScrollDownButtonTemplate")
	down:SetPoint("TOP",sb,"BOTTOM",0,8)
	down:SetScript("OnClick",function(self,button,down)
		PlaySound("UChatScrollButton")
		sb:SetValue(sb:GetValue() + sb:GetValueStep())
	end)

	---- border (why this was done by blizz in another frame I do not know)
	local border = CreateFrame("Frame",nil,sb)
	border:SetWidth(24)
	border:SetPoint("TOPLEFT",up,"TOPLEFT",-5,5)
	border:SetPoint("BOTTOMRIGHT",down,"BOTTOMRIGHT",5,-3)
	border:SetBackdrop({
		edgeFile="Interface\\Tooltips\\UI-Tooltip-Border",
		tile=true,
		edgeSize=12,
		tileSize=16,
		insets= {left=0, right=0, top=5, bottom=5},
	})
	border:SetBackdropBorderColor(TOOLTIP_DEFAULT_COLOR.r,TOOLTIP_DEFAULT_COLOR.g,TOOLTIP_DEFAULT_COLOR.b,0.5)
	
	-- Result List
	local list = CreateFrame("Frame",nil,rb)
	list:SetPoint("TOPLEFT"); list:SetPoint("BOTTOMLEFT");
	list:SetPoint("TOPRIGHT",sb,"TOPLEFT")
	list:SetPoint("BOTTOMRIGHT",sb,"BOTTOMLEFT")

	local head = CreateFrame("Frame",nil,list)
	head:SetPoint("TOPLEFT",list,"TOPLEFT",5,-2)
	head:SetPoint("TOPRIGHT")
	head:SetHeight(18)

	-- header buttons
	local name = CreateFrame("Button",nil,head,"OptionsButtonTemplate")
	name:SetPoint("TOPLEFT"); name:SetPoint("BOTTOMLEFT");
	name:SetWidth(120)
	name:SetText("Name")
	name:RegisterForClicks("LeftButtonUp")
	name:SetScript("OnClick",function(self,button,down) SocialistUI:ToggleNameHeader() end)

	local guild = CreateFrame("Button",nil,head,"OptionsButtonTemplate")
	guild:SetPoint("TOPLEFT",name,"TOPRIGHT")
	guild:SetPoint("BOTTOMLEFT",name,"BOTTOMRIGHT")
	guild:SetWidth(120)
	guild:SetText("Guild")
	guild:RegisterForClicks("LeftButtonUp")
	guild:SetScript("OnClick",function(self,button,down) SocialistUI:ToggleGuildHeader() end)

	local level = CreateFrame("Button",nil,head,"OptionsButtonTemplate")
	level:SetPoint("TOPLEFT",guild,"TOPRIGHT")
	level:SetPoint("BOTTOMLEFT",guild,"BOTTOMRIGHT")
	level:SetWidth(38)
	level:SetText("Lvl")
	level:RegisterForClicks("LeftButtonUp")
	level:SetScript("OnClick",function(self,button,down) SocialistUI:ToggleLevelHeader() end)

	local class = CreateFrame("Button",nil,head,"OptionsButtonTemplate")
	class:SetPoint("TOPLEFT",level,"TOPRIGHT")
	class:SetPoint("BOTTOMLEFT",level,"BOTTOMRIGHT")
	class:SetWidth(74)
	class:SetText("Class")
	class:RegisterForClicks("LeftButtonUp")
	class:SetScript("OnClick",function(self,button,down) SocialistUI:ToggleClassHeader() end)
	
	local entries = {}
	for i=1,MAXENTRIES do
		local entry = newResultEntry(list)
		if i==1 then
			-- First entry.
			entry:SetPoint("TOPLEFT",head,"BOTTOMLEFT",0,3)
			entry:SetPoint("TOPRIGHT",head,"BOTTOMRIGHT",0,3)
		else
			entry:SetPoint("TOPLEFT",entries[i-1],"BOTTOMLEFT")
			entry:SetPoint("TOPRIGHT",entries[i-1],"BOTTOMRIGHT")
		end
		entries[i] = entry
	end
	
	head.race = race
	head.level = level
	head.guild = guild
	head.name = name
	list.headers = head
	list.entries = entries
	sb.downArrow = down
	sb.upArrow = up
	rb.scrollBar = sb
	rb.resultList = list
	sf.resultBox = rb
end

---------------------------------
-- Pullout Frame               --
---------------------------------

do

	local po = CreateFrame("Frame",nil,sf)
	po:Hide()
	po:SetPoint("TOPLEFT",sf,"TOPRIGHT",0,-50)
	po:SetPoint("BOTTOMLEFT",sf,"BOTTOMRIGHT",0,50)
	po:SetWidth(200)
	po:SetBackdrop({bgFile = "Interface/Tooltips/UI-Tooltip-Background", 
		edgeFile = "Interface/Tooltips/UI-Tooltip-Border", 
		tile = true, tileSize = 16, edgeSize = 16, 
		insets = { left = 4, right = 4, top = 4, bottom = 4 },
		})
	po:SetBackdropColor(0,0,0,1)
	
	-- Title
	local tf = CreateFrame("Frame",nil,po)
	tf:SetPoint("TOPLEFT"); tf:SetPoint("TOPRIGHT");
	tf:SetHeight(24)
	local title = tf:CreateFontString(nil,"ARTWORK","GameFontNormal") --TODO: Better font
	title:SetAllPoints(tf)
	title:SetText("")
	po.title = title
	
	-- private variables
	
	po.lineRec = {} -- Frame Recycle Queue
	po.lineUse = {} -- Frame Active List
	po.lastAdded = tf
	
	-- private methods
	function po:NewFrame()
		local frame = nil
		if #(self.lineRec) > 0 then
			frame = tremove(self.lineRec)
		else
			frame = CreateFrame("Frame",nil,self)
		end
		tinsert(self.lineUse,frame)
		frame:Show()
		return frame
	end
	
	function po:AddHeader(info)
		local frame = self:NewFrame()
		frame:SetPoint("TOPLEFT",self.lastAdded,"BOTTOMLEFT")
		frame:SetPoint("TOPRIGHT",self.lastAdded,"BOTTOMRIGHT")
		frame:SetHeight(17)
		
		frame.text = frame.text or frame:CreateFontString(nil,"ARTWORK","GameFontNormal")
		local text = frame.text
		
		text:SetAllPoints(frame)
		text:SetJustifyH("CENTER")
		text:SetText(info)
		
		self.lastAdded = frame
	end
	
	function po:AddInfo(label,info)
		local frame = self:NewFrame()
		frame:SetPoint("TOPLEFT",self.lastAdded,"BOTTOMLEFT")
		frame:SetPoint("TOPRIGHT",self.lastAdded,"BOTTOMRIGHT")
		frame:SetHeight(17)
		
		frame.text = frame.text or frame:CreateFontString(nil,"ARTWORK","GameFontNormal")
		local text = frame.text
		
		text:SetAllPoints(frame)
		text:SetJustifyH("LEFT")
		text:SetText("   " ..label ..": ".. info)
		
		self.lastAdded = frame
	end
	
	function po:SetTitle(text)
		self.title:SetText(text)
	end
	
	function po:AddSpace()
		local frame = self:NewFrame()
		frame:SetPoint("TOPLEFT",self.lastAdded,"BOTTOMLEFT")
		frame:SetPoint("TOPRIGHT",self.lastAdded,"BOTTOMRIGHT")
		frame:SetHeight(17)
		
		frame.text = frame.text or frame:CreateFontString(nil,"ARTWORK","GameFontNormal")
		local text = frame.text
		
		text:SetAllPoints(frame)
		text:SetText("") -- Rather than making nil text, do a blank text
		-- This is so that we don't get a million FontStrings sitting around doing nothing
		-- in memory
		
		self.lastAdded = frame
	end
	
	function po:Clear()
		self.lastAdded = tf
		while #(self.lineUse) > 0 do
			local frame = tremove(po.lineUse)
			frame:Hide()
			tinsert(po.lineRec,frame)
		end
	end
	
	-- Access Methods
	
	function po:SetEntry(entry)
		self:Clear()
		-- Entry comes directly from the database
		self.entry = entry
		
		if entry then
			self:SetTitle(entry.name)

			-- Info
			self:AddHeader("Information")
			if entry.guild then self:AddInfo("Guild",entry.guild) end
			if entry.level then self:AddInfo("Level",entry.level) end
			if entry.class then self:AddInfo("Class",entry.class) end
			if entry.race then self:AddInfo("Race",entry.race) end

			self:AddSpace()

			if entry.places then
				self:AddHeader("Last Seen")
				for k,v in ipairs(entry.places) do
					self:AddInfo(date("%m/%d/%y",v["date"]),v["place"])
				end
			end
		else
			self:SetTitle("")
		end
	end
	
	function po:GetEntry()
		return self.entry
	end
	
	sf.pullout = po
end

---------------------------------
-- VARIABLES                   --
---------------------------------

SocialistUI.topResultIndex = 1
SocialistUI.orderMethod = "NAME"

---------------------------------
-- METHODS                     --
---------------------------------

function SocialistUI:SetResult(newresult)
	self.result = newresult
	if #newresult < MAXENTRIES then
		self.resultBox.scrollBar:SetMinMaxValues(1,1)
	else
		self.resultBox.scrollBar:SetMinMaxValues(1,#newresult - MAXENTRIES)
	end
	self:Order()
	self:RedrawList()
end

--- COMPARISON METHODS (Lua uses 'quicksort')
-- Yes, there are ways to bring these down to less or even 1 sort function,
-- but I prefer to keep them split as such for future extensibility.
SocialistUI.nameSort = function(a,b)return a.name < b.name end
SocialistUI.nameRSort = function(a,b) return a.name > b.name end
SocialistUI.classSort = function(a,b)
	if not a.class and b.class then return false end
	if not b.class and a.class then return true end
	if not a.class and not b.class then return SocialistUI.nameSort(a,b) end
	if a.class == b.class then return SocialistUI.nameSort(a,b) end
	return a.class < b.class
end
SocialistUI.classRSort = function(a,b)
	if not a.class and b.class then return true end
	if not b.class and a.class then return false end
	if not a.class and not b.class then return SocialistUI.nameRSort(a,b) end
	if a.class == b.class then return SocialistUI.nameRSort(a,b) end
	return a.class > b.class
end
SocialistUI.guildSort = function(a,b)
	if (not a.guild or a.guild == "") and (b.guild and b.guild ~= "")then return false end
	if (not b.guild or b.guild == "") and (a.guild and a.guild ~= "") then return true end
	if (not a.guild or a.guild == "") and (not b.guild or b.guild == "") then return SocialistUI.nameSort(a,b) end
	if a.guild == b.guild then return SocialistUI.nameSort(a,b) end
	return a.guild < b.guild
end
sf.guildRSort = function(a,b)
	if (not a.guild or a.guild == "") and (b.guild and b.guild ~= "")then return true end
	if (not b.guild or b.guild == "") and (a.guild and a.guild ~= "") then return false end
	if (not a.guild or a.guild == "") and (not b.guild or b.guild == "") then return SocialistUI.nameRSort(a,b) end
	if a.guild == b.guild then return SocialistUI.nameRSort(a,b) end
	return a.guild > b.guild
end
SocialistUI.levelSort = function(a,b)
	if not a.level and b.level then return false end
	if not b.level and a.level then return true end
	if not a.level and not b.level then return SocialistUI.nameSort(a,b) end
	if a.level == b.level then return SocialistUI.nameSort(a,b) end
	return a.level > b.level
end
SocialistUI.levelRSort = function(a,b)
	if not a.level and b.level then return true end
	if not b.level and a.level then return false end
	if not a.level and not b.level then return SocialistUI.nameRSort(a,b) end
	if a.level == b.level then return SocialistUI.nameRSort(a,b) end
	return a.level < b.level
end

function SocialistUI:Order()
	-- TODO: Add new sorting methods.
	if self.orderMethod == "GUILD" then
		table.sort(self.result,self.guildSort)
	elseif self.orderMethod == "GUILDR" then
		table.sort(self.result,self.guildRSort)
	elseif self.orderMethod == "LEVEL" then
		table.sort(self.result,self.levelSort)
	elseif self.orderMethod == "LEVELR" then
		table.sort(self.result,self.levelRSort)
	elseif self.orderMethod == "CLASS" then
		table.sort(self.result,self.classSort)
	elseif self.orderMethod == "CLASSR" then
		table.sort(self.result,self.classRSort)
	elseif self.orderMethod == "NAMER" then
		table.sort(self.result,self.nameRSort)
	else -- "NAME" or error
		table.sort(self.result, self.nameSort)
	end
end

function SocialistUI:ToggleNameHeader()
	if self.orderMethod == "NAME" then
		self.orderMethod = "NAMER"
	else
		self.orderMethod = "NAME"
	end
	self:Order()
	self:RedrawList()
end

function SocialistUI:ToggleGuildHeader()
	if self.orderMethod == "GUILD" then
		self.orderMethod = "GUILDR"
	else
		self.orderMethod = "GUILD"
	end
	self:Order()
	self:RedrawList()
end

function SocialistUI:ToggleLevelHeader()
	if self.orderMethod == "LEVEL" then
		self.orderMethod = "LEVELR"
	else
		self.orderMethod = "LEVEL"
	end
	self:Order()
	self:RedrawList()
end

function SocialistUI:ToggleClassHeader()
	if self.orderMethod == "CLASS" then
		self.orderMethod = "CLASSR"
	else
		self.orderMethod = "CLASS"
	end
	self:Order()
	self:RedrawList()
end

-- RedrawList() -
--  Redraw the current elements of the resultList based on self.result
--  DO NOT assume that self.result is set
--  DO NOT assume that any result elements are set,unset,hidden,shown
function SocialistUI:RedrawList()
	if self.result then
		for enum=1,MAXENTRIES do
			if self.topResultIndex + enum - 1 > #self.result then
				-- Note that this will only happen in the case that there are less entries in the result
				-- than MAXENTRIES
				self:ClearEntry(enum)
			else
				self:SetEntry(enum,self.result[self.topResultIndex + enum - 1])
			end
		end
	else
		self.topResultIndex = 1
		for enum=1,MAXENTRIES do
			self:ClearEntry(enum)
		end
	end
end

function SocialistUI:ClearEntry(i)
	-- I don't think clearing anything is actually necessary. Hiding is sufficient.
	sf.resultBox.resultList.entries[i]:Hide()
end

function SocialistUI:SetEntry(i,resultEntry)
	local box = sf.resultBox.resultList.entries[i]
	box.name.text:SetText(resultEntry.name or "ERROR")
	box.guild.text:SetText(resultEntry.guild or "")
	if not resultEntry.level or resultEntry.level < 1 then
		box.level.text:SetText("")
	else
		box.level.text:SetText(resultEntry.level)
	end
	box.class.text:SetText(resultEntry.class or "")
	box:Show()
end

function SocialistUI:ClearAll()
	for i=1,MAXENTRIES do
		self:ClearEntry(i)
	end
end

function SocialistUI:ReturnToTop()
	self.topResultIndex = 1
	self:RedrawList()
end

function SocialistUI:TextChanged()
	if not self.updating then
		self.more = false
		self.updating = true
		sf:SetScript("OnUpdate", self.FormOnUpdate)
		self.lastupdate = 0
	else
		self.more = true
	end
end

function sf.FormOnUpdate(self, elapsed)
	self = SocialistUI -- Hack. Yes there are better ways, but this kept changes to modularity simple.
	self.lastupdate = self.lastupdate + elapsed
	if (self.lastupdate >= MIN_TEXT_UPDATE_INTERVAL) then
		self:ProcessQuery(self:GetQuery())
		if self.more then
			-- There is more waiting for us, go work.
			self.more = false
			self.lastupdate = 0
		else
			self.updating = false
			sf:SetScript("OnUpdate", nil)
		end
	end
end

function SocialistUI:GetQuery()
	-- Return a table based on the query.
	-- TODO: formalize the meanings of nil, empty strings, etc.
	local query = {}
	query.name = sf.filterBox.nameBox:GetText()
	if query.name == "" then query.name = nil end
	query.guild = sf.filterBox.guildBox:GetText()
	if query.guild == "" then query.guild = nil end
	local minLevel = sf.filterBox.levelMinBox:GetText()
	local maxLevel = sf.filterBox.levelMaxBox:GetText()
	if minLevel then minLevel = tonumber(minLevel) end
	if maxLevel then maxLevel = tonumber(maxLevel) end
	query.minLevel = minLevel
	query.maxLevel = maxLevel
	return query
end

function SocialistUI:ProcessQuery(query)
	-- Process the SocialistForm query. Called by text change or button press.
	-- needs to be as fast as possible! Extreme optimization. Called often.
	-- Return nothing. Write to frame.

    local result = {}
    local itor = 1
    for k,v in pairs(Socialist:Query(query,Socialist.db.realm.listing) or {}) do
        result[itor] = v
        v.name = k
        itor = itor + 1
    end
    self:SetResult(result)
end

function SocialistUI:ResultClick(result)
	local entry = Socialist:GetEntry(result.name.text:GetText())
	if sf.pullout:GetEntry() == entry then -- Don't confuse pullot:GetEntry with Socialist:GetEntry
		if sf.pullout:IsShown() then sf.pullout:Hide() else sf.pullout:Show() end
	else
		sf.pullout:Show()
		sf.pullout:SetEntry(entry)
	end

end

function SocialistUI:QueryServer()
	local query = self:GetQuery()
	wholib:Who(Socialist:FormulateQuery(query.name,query.guild,nil,nil,nil,query.minLevel,query.maxLevel))
end

function SocialistUI:UpdateList()
	self:ProcessQuery(self:GetQuery())
end


function sf:ToggleMainForm()
		if sf:IsShown() then
			sf:Hide()
		else
			sf:Show()
		end

end
